/*
 *  ofxiPhoneVideoPlayer.cpp
 *  ffmpeg
 *
 *  Created by theo on 13/12/2009.
 *  Copyright 2009 __MyCompanyName__. All rights reserved.
 *
 */

#include "ofxiPhoneVideoPlayer.h"

//most of this code comes from http://www.gamedev.net/community/forums/topic.asp?topic_id=552132

ofxiPhoneVideoPlayer::ofxiPhoneVideoPlayer(void){
	bActive			= true;
	data			= 0;
	dwNext			= 0;
	dNext			= 0;
	width			= 0;
	height			= 0;
	nwidth			= 0;
	nheight			= 0;
	dwFrame			= 0;
	dwFrameSys		= 0;
	dPts			= 0.0;
	bLoad			= false;
	bLoop			= true;
	bResize			= false;	
	millisLastFrame = 0;
	
	//theo 
	pixels		= NULL;
	bUseTexture = true;
	bPlaying    = false;
}

ofxiPhoneVideoPlayer::~ofxiPhoneVideoPlayer(void){
	CloseAVI();
	if( pixels!=NULL ){
		delete[] pixels;
		pixels = NULL;
	}
}

bool ofxiPhoneVideoPlayer::loadMovie(string filename){
	bLoaded = OpenAVI(filename);
	millisLastFrame = ofGetElapsedTimeMillis();
	
	if( bLoaded && bUseTexture ){
		tex.allocate(width, height, GL_RGB);

		if( pixels!=NULL ){
			delete[] pixels;
			pixels = NULL;
		}
		pixels   = new unsigned char[ width * height * 3 ];

		//get some pixels in there
		bPlaying = true;
		update();
		bPlaying = false;
			
		SetZeroCount();
	}
	
	return bLoaded;
}

void ofxiPhoneVideoPlayer::play(){
	bPlaying = true;
}

void ofxiPhoneVideoPlayer::stop(){
	bPlaying = false;
}

void ofxiPhoneVideoPlayer::setPaused(bool bPaused){
	bPlaying = !bPaused;
}

void ofxiPhoneVideoPlayer::update(){
	int millis = ofGetElapsedTimeMillis(); 

	if( bLoaded && bPlaying ){
		
		int deltaMillis = millis - millisLastFrame;
		deltaMillis = ofClamp(deltaMillis, 1, 20000);
		
		Update(deltaMillis);
		GrabAVIFrame();
		
		if(frameFinished ){
			
//			pixels = buffer;//
			memcpy( pixels, buffer, width*height*3);

//			unsigned char * src = &buffer[0];
//			unsigned char * dst = &pixels[(width*(height-1)*3)];
//			int lineSize = width * 3;
//			
//			for(int k = 0; k < height; k++){
//				memcpy( dst, src, lineSize);
//				(src)+=lineSize;
//				(dst)-=lineSize;
//			}
//				
			if( bUseTexture ){
				//printf("loading pixels into texture\n");
				tex.loadData(pixels, width, height, GL_RGB);
			}
		}
	}
	
	millisLastFrame = millis;
}

unsigned char * ofxiPhoneVideoPlayer::getPixels(){

	//should be return pixels - but for some reason the pixels are coming out flipped
	//if we don't flip pixels in update() then the texture is flipped - weird!
	return pixels;
}

//------------------------------------
void ofxiPhoneVideoPlayer::draw(float _x, float _y, float _w, float _h){
	if (bUseTexture){
		tex.draw(_x, _y+_h, _w, -_h);
	}
}

//------------------------------------
void ofxiPhoneVideoPlayer::draw(float _x, float _y){
	draw(_x, _y, (float)width, (float)height);
}




//protected stuff

bool ofxiPhoneVideoPlayer::OpenAVI(string szFile)
{

 // must be called before using avcodec lib 
    avcodec_init();

    // register all the codecs 
    av_register_all();

	chFile = ofToDataPath(szFile);
	
	if(!CreateContext()) return 0;
	pFrameRGB=avcodec_alloc_frame();
    if(pFrameRGB==NULL)
	{
		printf("FFMPEG: Can't Alloc Frame! \"%s\"\n", chFile.c_str());
		DestroyContext();
        return 0;
	}

	nwidth = pCodecCtx->width;//ofNextPow2(pCodecCtx->width);
	nheight = pCodecCtx->height;//ofNextPow2(pCodecCtx->height);

	width  = pCodecCtx->width;
	height = pCodecCtx->height;

    numBytes=avpicture_get_size(PIX_FMT_RGB24, nwidth, nheight);
    buffer=new uint8_t[numBytes];

    avpicture_fill((AVPicture *)pFrameRGB, buffer, PIX_FMT_RGB24, nwidth, nheight);

	SetResize();
	dwFrame=0;
	dwFrameSys = 0;
	dwNext=0;
	dNext=0;
	dPts=0;
	bLoad = true;
	bActive = true;
	return true;
}

#define MAX_VIDEOQ_SIZE (5 * 256 * 1024)

void ofxiPhoneVideoPlayer::GrabAVIFrame()
{
	if(!bActive) return;
	int nDifFrame = dwFrame - dwFrameSys;
	if(nDifFrame > 0)
	{
		for(int k = 0; k < nDifFrame; k++)
		{
			if(av_read_frame(pFormatCtx, &packet)<0) 
			{
				bActive = false;
				break;
			}
			if(k != (nDifFrame-1)) av_free_packet(&packet);
		}
		if(!bActive) 
		{
			if(bLoop) SetZeroCount();
			return;
		}
		if(packet.stream_index==videoStream)
		{
			dPts = 0;
			// Decode video frame
			avcodec_decode_video(pCodecCtx, pFrame, &frameFinished, 
				packet.data, packet.size);

			if(packet.dts == AV_NOPTS_VALUE 
			   && pFrame->opaque && *(uint64_t*)pFrame->opaque != AV_NOPTS_VALUE) 
			{
				dPts = double(*(uint64_t *)pFrame->opaque);
			} 
			else if(packet.dts != AV_NOPTS_VALUE) 
			{
				dPts = double(packet.dts);
			} 
			else 
			{
				// No se sabe el tiempo, entonces se asigna al tiempo actual
				dPts = 0.0;
			}
			dPts *= av_q2d(video_st->time_base);
			if(dPts == 0.0) dPts = dNext;

			// Did we get a video frame?
			if(frameFinished)
			{
				sws_scale (img_convert_ctx, pFrame->data, pFrame->linesize,
							0, pCodecCtx->height,
							pFrameRGB->data,pFrameRGB->linesize);
				data = pFrameRGB->data[0];
			}
		}
		av_free_packet(&packet);
	}
}

int ofxiPhoneVideoPlayer::Update(DWORD milliseconds)
{
	dwNext += milliseconds;
	dNext = double(dwNext)/1000;	// Convertir a segundos
	dwFrame = DWORD(dNext/dTimeBase);
	dwFrameSys = DWORD(dPts/dTimeBase);
	return 0;
}
void ofxiPhoneVideoPlayer::SetZeroCount(void)
{
	dwFrame=0;
	dwFrameSys = 0;
	dwNext=0;
	dNext=0;
	dPts=0;
	DestroyContext();
	CreateContext();
	bActive = true;
}

void ofxiPhoneVideoPlayer::CloseAVI(void)
{
	if(bLoad) 
	{
		DestroyContext();
		delete [] buffer;
		if(pFrameRGB != NULL) av_free(pFrameRGB);
	}
	bLoad = false;
	if(bResize) ClearResize();
}

void ofxiPhoneVideoPlayer::SetResize(void)
{
	static int sws_flags = SWS_BICUBIC;
	img_convert_ctx = sws_getContext(pCodecCtx->width, 
									pCodecCtx->height,
									pCodecCtx->pix_fmt,
									nwidth, 
									nheight,
									PIX_FMT_RGB24,
									sws_flags, NULL, NULL, NULL);
	bResize = true;
}

void ofxiPhoneVideoPlayer::ClearResize(void)
{
	sws_freeContext(img_convert_ctx);
	bResize = false;
}

// ** Funciones privadas
int ofxiPhoneVideoPlayer::CreateContext(void)
{
	int result = av_open_input_file(&pFormatCtx, chFile.c_str(), NULL, 0, NULL);
	if(result!=0)
	{
		printf("FFMPEG: Can't open file! %i \"%s\"\n", result, chFile.c_str());
        return 0;
	}

    if(av_find_stream_info(pFormatCtx)<0)
	{
		printf("FFMPEG: Couldn't find stream information! \"%s\"\n", chFile.c_str());
        return 0;
	}

    dump_format(pFormatCtx, 0, chFile.c_str(), false);
    videoStream=-1;
	int i;
    for(i=0; i<int(pFormatCtx->nb_streams); i++)
        if(pFormatCtx->streams[i]->codec->codec_type==CODEC_TYPE_VIDEO)
        {
            videoStream=i;
            break;
        }
    if(videoStream==-1)
	{
		printf("FFMPEG: Didn't find a video stream! \"%s\"\n", chFile.c_str());
        return 0; 
	}

    // Get a pointer to the codec context for the video stream
	video_st=pFormatCtx->streams[videoStream];
    pCodecCtx=pFormatCtx->streams[videoStream]->codec;

    pCodec=avcodec_find_decoder(pCodecCtx->codec_id);
    if(pCodec==NULL)
	{
		printf("FFMPEG: Codec not found! \"%s\"\n", chFile.c_str());
        return 0;
	}

    if(avcodec_open(pCodecCtx, pCodec)<0)
	{
		printf("FFMPEG: Could not open codec! \"%s\"\n", chFile.c_str());
        return 0;
	}

    pFrame=avcodec_alloc_frame();

	dTimeBase = av_q2d(video_st->time_base);
	dDur = double(video_st->duration)*dTimeBase;
	dwDur = DWORD(dDur*1000);

	GrabAVIFrame();

	return 1;
}

void ofxiPhoneVideoPlayer::DestroyContext(void)
{
    av_free(pFrame);
    avcodec_close(pCodecCtx);
    av_close_input_file(pFormatCtx);
}

